/* generated by generate.py */
#![allow(dead_code)]
#![allow(non_camel_case_types)]
#![allow(non_snake_case)]
#![allow(non_upper_case_globals)]
pub mod virgl_debug_callback_bindings;
pub mod virgl_renderer_bindings;
